/* Copyright 1998 by Scott Franke [druid-]
    sfranke@scf.usc.edu
    http://www-scf.usc.edu/~sfranke/glj
*/	

#ifndef CONSOLE_H
#define CONSOLE_H

#define VK_PAGEUP   33
#define VK_PAGEDOWN 34

#define CONSOLE_MAX_ROW 60

static float version = 0.10f;

/* CONSOLE STRUCTURES */

typedef struct {
	unsigned char r;
	unsigned char g;
	unsigned char b;
} color_t;

typedef struct {
	unsigned char flash;
	color_t color;
	char text[60];
} row_t;

typedef void (*cmd_ptr)(struct console_s *, char *);

typedef struct cmd_s {
	char fullcmd[15];
	void (*function)(struct console_s *, char *);
	struct cmd_s *next;
} cmd_t;

typedef struct console_s {
	row_t buffer[CONSOLE_MAX_ROW];
	int newline;
	int col;
	int viewbottom;

	char cmdhist[30][60];
	int newcmd,numcmdhist,curcmd;

	char ver[10];
	BOOL drawortho;

	cmd_t *cmds[256];

	int font_w, font_h;		// characters in a row, column

	int active, destroying, creating;
	int font_id, logo_id;

} console_t;

	// Mode the main code is in
extern int mode;
#define MODE_CONSOLE		0
#define MODE_PLAYBACK		1
#define MODE_INTERACTIVE	2

	// Signifying to quit the program
extern int done;

	// color for any new text
color_t console_color;

	/* Console functions */
void GLConInitConsole( void );
console_t *GLConCreateConsole( float, char *, char * );
void GLConDestroy( console_t * );
void GLConAddCmd( console_t *, char *, void (*function)(console_t *, char *) );

void GLConExecFile( console_t *, char * );

void GLConDrawConsole(console_t *, float );

void GLConIn( console_t *, long );
void GLConCmdHistory( console_t *, DWORD );
void GLConScroll( console_t *, DWORD );

void GLConAdd( console_t *, char * );
void GLConOut( console_t *, char * );
void GLConAlert( console_t *, char * );
void GLConAdvance( console_t * );

#endif