#include <windows.h>
#include <GL/gl.h>
#include <stdio.h>
#include "tex.h"

extern BOOL ext_pal_support;
static PFNGLCOLORTABLEEXTPROC glColorTable;

BOOL GLSetupPalSupport( void )
{
	if( !ext_pal_support )
		return FALSE;
	glColorTable = (PFNGLCOLORTABLEEXTPROC) wglGetProcAddress( "glColorTableEXT" );
	if( !glColorTable )
		return FALSE;
	return TRUE;
}

	//
	// Load an 8 bit BMP image
	//
void GLLoadBMP8 (char *filename, int tex_id, int tex_mode )
{
	BMPheader bh;
	BMPinfo bp;
	BMPcoreinfo bc;
	FILE *fp;
	BOOL error = FALSE;
	byte paldata[1024], *image;
	char errorStr[30];
	int imagesize;
	int i, j, x, y;

	sprintf( errorStr, filename );

	fp = fopen( filename, "rb" );

	if( !fp )
	{
		strcat( errorStr, "\nUnable to open file" );
		error = TRUE;
	}

	if( !error )
	{

		if(fread(&bh.type, sizeof(BMPheader)-2, 1, fp) != 1)
		{
			strcat( errorStr, "\nFailed to read bitmap header" );
			error = TRUE;
		}

		if (bh.strsize == 40)
		{
				// Infoheader
			fread(&bp.width, sizeof(BMPinfo), 1, fp);
			if (bp.planes != 1)
			{
				strcat( errorStr, "\nPlanes wrong" );
				error = TRUE;
			}
			if (bp.compression != 0)
			{
				strcat( errorStr, "\nCompression wrong" );
				error = TRUE;
			}
			if (bp.bitcount != 8)
			{
				strcat( errorStr, "\nNot 8-bit bitmap" );
				error = TRUE;
			}
			if (bp.imagesize)
			{
				if (bp.imagesize != (bp.width*bp.height))
				{
					strcat( errorStr, "\nImage size incorrect" );
					error = TRUE;
				}
			}
			imagesize = bp.width*bp.height;
			x = bp.width;
			y = bp.height;

			if (bp.clrused)
				fread(paldata, 4*bp.clrused, 1, fp);
			else
				fread(paldata, 1024, 1, fp);
		
		}

	}
	else if (bh.strsize == 12)
	{
			// core infoheader
		fread(&bc.width, sizeof(BMPcoreinfo), 1, fp);
			if (bc.planes != 1)
			{
				strcat( errorStr, "\nPlanes wrong" );
				error = TRUE;
			}
			if (bc.bitcount != 8)
			{
				strcat( errorStr, "\nNot 8-bit bitmap" );
				error = TRUE;
			}
			imagesize = bc.width*bc.height;
			x = bc.width;
			y = bc.height;

			fread(paldata, 256*3, 1, fp);
	}
	else
	{
		strcat( errorStr, "\nInvalid bitmap format" );
		error = TRUE;
	}

	if( !error )
	{
		if ( !(image = malloc(imagesize)) )
		{
			strcat( errorStr, "\nMalloc failed in GLLoadBMP8" );
			error = TRUE;
		}
		if( !error )
		{
			fread(image, imagesize, 1, fp);
		}
	}
	if( fp )
		fclose( fp );
	if( error )
	{
		MessageBox( NULL, errorStr, "Error loading texture", MB_OK|MB_ICONERROR );
	}
	else
	{
			// ARGB to RGBA
		for( i=0;i<256;i++ )
		{
			static unsigned char r,g,b,a;
			r = paldata[i*4+2];
			g = paldata[i*4+1];
			b = paldata[i*4+0];
			a = paldata[i*4+3];
			paldata[i*4] = r;
			paldata[i*4+1] = g;
			paldata[i*4+2] = b;
			paldata[i*4+3] = a;
		}

		switch( tex_mode )
		{
		case TEX_ALPHA_COLORKEY_0:
				// "colorkey'd" to entry 0
			paldata[3] = 0;
			for( i=1;i<256;i++ )
				paldata[i*4+3] = 255;
			break;
		case TEX_ALPHA_HIGHRGB:
			for( i=1;i<256;i++ )
			{
				paldata[i*4+3] = paldata[i*4];
				if( paldata[i*4+1] > paldata[i*4+3] )
					paldata[i*4+3] = paldata[i*4+1];
				if( paldata[i*4+2] > paldata[i*4+3] )
					paldata[i*4+3] = paldata[i*4+2];
			}
			break;
		}

		glBindTexture(GL_TEXTURE_2D, tex_id);
		if( ext_pal_support )
		{
			glColorTable ( GL_TEXTURE_2D,
				GL_RGBA,
				256,
				GL_RGBA,
				GL_UNSIGNED_BYTE,
				paldata );
			glTexImage2D( GL_TEXTURE_2D,
				0,
				GL_COLOR_INDEX8_EXT,
				x,
				y,
				0,
				GL_COLOR_INDEX,
				GL_UNSIGNED_BYTE,
				image );
		}
		else
		{
			char *fullimage;
			fullimage = (char *) malloc(x*y*4);
			for( i=0;i<y;i++ )
				for( j=0;j<x;j++ )
				{
					fullimage[i*x*4+j*4]   = paldata[image[i*x+j]*4];
					fullimage[i*x*4+j*4+1] = paldata[image[i*x+j]*4+1];
					fullimage[i*x*4+j*4+2] = paldata[image[i*x+j]*4+2];
					fullimage[i*x*4+j*4+3] = paldata[image[i*x+j]*4+3];
				}
			glTexImage2D( GL_TEXTURE_2D,
				0,
				GL_RGBA,
				x,
				y,
				0,
				GL_RGBA,
				GL_UNSIGNED_BYTE,
				fullimage );
			if( fullimage )
				free( fullimage );
		}
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
		glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	}
	if( image )
		free( image );
}

	//
	// Load a 24-bit RAW image
	//
void GLLoadRAW (char *filename, int tex_id, int x, int y)
{
 FILE *FilePic;
 long i,j,k;

 if ((x > 256) || (y > 256))
 {
	MessageBox(NULL, "Texture too large", "Error Loading Texture", MB_ICONSTOP);
	exit (1);
 }
 if ((FilePic=fopen(filename,"r"))==NULL)
 {
	MessageBox(NULL, filename, "Texture Not Found", MB_ICONSTOP);
	exit (1);
 }
 fread (bufferpic , x*y*3, 1 , FilePic);
 fclose (FilePic);

 for (i=0;i<x;i++)
   {
    for (j=0;j<x;j++)
    { pic[i][j][0]=bufferpic[i][j][0];
	 	pic[i][j][1]=bufferpic[i][j][1];
	 	pic[i][j][2]=bufferpic[i][j][2];

		pic[i][j][3]=pic[i][j][0];
		for(k=1;k<3;k++)
		if((pic[i][j][k]) > (pic[i][j][3]))
			pic[i][j][3] = pic[i][j][k];
    }
   }

	glBindTexture(GL_TEXTURE_2D, tex_id);
	glTexImage2D(GL_TEXTURE_2D, 0, 4, 256, 256, 0, GL_RGBA,
		     GL_UNSIGNED_BYTE, &pic[0][0][0]);

	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
}
