#define CORE

#include <windows.h>
#include <process.h>
#include "driver.h"

//** Externals
BOOL DI_Init(void);
void DI_End(void);
void RunCPU();

//** Gloabls
unsigned int hMainThread;

//** Prototypes
void RemoveCardridge();
void ResetCPU();
void ResumeCPU();
void StartRom();
void StopCPU();
void PauseCPU();

void InitDrivers(HWND aWindow,HINSTANCE anInstance);
void ReleaseDrivers();

void ExecuteThread(void*p);
void StartCPU(char *RomName);
void cpu_reset();


extern BOOL CPUIsRunning;

//** DirectInput
extern BOOL InitDirectInput(HWND hWindow, HINSTANCE hInstance);
extern void ReleaseDirectInput();

//** functions
void RemoveCardridge()
{
// Clean Up CPU 

}

void ResetCPU()
{
	// FIXME
	cpu_reset(); // in cpu.c
}


void PauseCPU()
{
	SuspendThread((HANDLE)hMainThread);
}

void ResumeCPU()
{
	ResumeThread((HANDLE)hMainThread);
}

extern int Init_Emulator(char *RomName);


#include "Global/type_sizes.h"
#include "Global/memory_extern.h"
#include "n64/registers.h"



void StopCPU()
{
	CPUIsRunning = FALSE;
	ShowWindow(DisplayWinData.hWnd, 0);
//	freeCompiler();
//	freeMemoryArena();
}

void InitDrivers(HWND aWindow,HINSTANCE anInstance)
{
	if(LoadVideoDriver("Video.dll") != 0)
	{
		MessageBox(NULL, "one", "", MB_OK);
		exit(-1);
	}
	if(LoadAudioDriver("Sound.dll") != 0)
	{
		MessageBox(NULL, "two", "", MB_OK);
		exit(-1);
	}
}

void ReleaseDrivers()
{
	ReleaseVideoDriver();
	ReleaseAudioDriver();
	ReleaseDirectInput();
}


extern HINSTANCE hVideoDriver;
void StartRom(char *RomName)
{
	tCoreConnection	aConnection;

	if (Init_Emulator(RomName) == 1)
	{
		aConnection.pIMEM = (unsigned char *)mem.sp_imem;
		aConnection.pDMEM = (unsigned char *)mem.sp_dmem;
		aConnection.pRDRAM = (unsigned char *)mem.rd_ram;
		aConnection.pVIREG = (unsigned char *)mem.vi_reg;
		aConnection.pInterruptMask = &reg.do_or_check_sthg;
		Vid_CreateConnectionWithCore(&aConnection);
		Aud_CreateConnectionWithCore(&aConnection);

		ShowWindow(DisplayWinData.hWnd, 1);
		CPUIsRunning = TRUE;

		hMainThread = _beginthread(ExecuteThread,4096,NULL);
	}
}

void ExecuteThread(void *p)
{	
	Vid_InitVideo(DisplayWinData.hWnd, hVideoDriver);	
	Aud_InitAudio(DisplayWinData.hWnd, hVideoDriver);
	InitDirectInput(DisplayWinData.hWnd, hVideoDriver);

	while(CPUIsRunning) RunCPU();

	Vid_ReleaseVideo();
	Aud_ReleaseAudio();
	ReleaseDirectInput();
}
