#include <windows.h>
#include <GL/gl.h>
#include <GL/glu.h>
#include <stdio.h>
#include <stdlib.h>

#include "rdp_registers.h"
#include "texture.h"

//** externals
extern int				*pInterruptMask;
extern unsigned char	*pRDRAM;
extern unsigned char	*pIDMEM;
extern unsigned char	*pVIREG;
extern HWND hGraphics;
extern int ucode_version;

//** texture - caching
#define CACHE_TEXTURES
#define MAX_CACHED_TEXTURES 4096
//#define MAX_CACHED_TEXTURES 512
static t_old_texture old_texture[MAX_CACHED_TEXTURES];
static t_index_texture index_texture[MAX_CACHED_TEXTURES];
unsigned int num_cached_texture = 0;

//** global variables
static unsigned char tbuffer0[256*256*4];
static unsigned char tbuffer1[256*256*4];
static unsigned char tbuffer2[256*256*4];

int TexMode = 0;
uWORD TexColor = 0;

//** prototypes
static int TestPower2(int width, int height, int *nw, int *nh);
static void CopyTextureRGBA32(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureRGBA16(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureYUV   (uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureCI8   (uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureCI4   (uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureIA16  (uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureIA8   (uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureIA4   (uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureI16   (uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureI8    (uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);
static void CopyTextureI4    (uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height);

static void NCopyTextureRGBA32(uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
static void NCopyTextureRGBA16(uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
static void NCopyTextureCI8   (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
static void NCopyTextureCI4   (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
static void NCopyTextureIA16  (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
static void NCopyTextureIA8   (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
static void NCopyTextureIA4   (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
static void NCopyTextureI16   (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
static void NCopyTextureI8    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
static void NCopyTextureI4    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);
//static void NCopyTextureYUV   (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey);

/******************************************************************************\
*                                                                              *
*   helper routines									  						   *
*                                                                              *
\******************************************************************************/

uWORD Palette8[256];
/*
static int TestPower2(int width, int height, int *nw, int *nh)
{
	int w;
	int temp, NewWidth = *nw, NewHeight = *nh;
	
//** test if width has the power of 2
	temp = width;
	for (w =11; w>= 0; w--)
	{
		if(temp&0x800)
		{
			NewWidth = (1<<w);
			break;
		}
		temp<<=1;
	}
//** test if height has the power of 2
	temp = height;
	for (w =11; w>= 0; w--)
	{
		if(temp&0x800)
		{
			NewHeight = (1<<w);
			break;
		}
		temp<<=1;
	}
//** return
	if ((NewWidth == width) && (NewHeight == height)) return 1;
	else
	{
		*nw = NewWidth;
		*nh = NewHeight;
		return 0;
	}
	return 1;
}
*/
static int TestPower2(int width, int height, int *nw, int *nh)
{
	int w;
	int temp, NewWidth = *nw, NewHeight = *nh;
	
//** test if width has the power of 2
	temp = 1;
	for (w = 0; w <= 9;w++)
	{
		NewWidth = temp;
		temp = temp << 1;
		if (NewWidth >= width)
		{
			break;
		}
	}
//** test if height has the power of 2
	temp = 1;
	for (w = 0; w <= 9;w++)
	{
		NewHeight = temp;
		temp = temp << 1;
		if (NewHeight >= height)
		{
			break;
		}
	}
//** return
	if ((NewWidth == width) && (NewHeight == height)) return 1;
	else
	{
		*nw = NewWidth;
		*nh = NewHeight;
		return 0;
	}
	return 1;
}

/******************************************************************************\
*                                                                              *
*   Texture Handling								  						   *
*                                                                              *
\******************************************************************************/
int twidth,tstart;
extern int TextWidth;

int DecompressTexture(t_tile *tmpTile,unsigned char *textaddr,int select)
{
//	unsigned char *src  = &pRDRAM[rdp_reg.TextureImage.addr];
	unsigned char *src  = textaddr;
	unsigned char *dest;

    int width  = tmpTile->lrs - tmpTile->uls;// + 1;
    int height = tmpTile->lrt - tmpTile->ult;// + 1;

	uBYTE size   = tmpTile->size;
	uBYTE format = tmpTile->format;

	if (select == 0)
		dest = &tbuffer1;
	else
		dest = &tbuffer0;

//    width  = tmpTile->Width;
	twidth = rdp_reg.TextureImage.width;
//	twidth = tmpTile->line;

	if (width < 0) width = - width;
	if (height < 0) height = - height;

	width++;
	height++;

	if (twidth < 0) twidth = -twidth;

	if (twidth <= 2)
		twidth = tmpTile->lrs + 1;//width;

	tstart = width & 1;
	if ((width > 350) || (twidth > 350) || (height > 350))
	{
		return 0;
	}

	switch(format)
	{
	case 0:	//rgba // street
			switch(size)
			{
			case 0:
					if (twidth < (width-2)) twidth += twidth;
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureCI4(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
					break;
			case 2:
					//twidth = (twidth) & 0xfffe;
//					CopyTextureRGBA16(src, dest, width, height);
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureRGBA16(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
					break;
			case 3:
//				    CopyTextureRGBA32(src, dest, width, height);
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureRGBA32(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
					break;
			}			
			break;
	case 1:	
			//**not be implemented
			CopyTextureYUV(src, dest, width, height);
			break;
	case 2:	//CI
			switch(size)
			{
			case 0:
//				    CopyTextureCI4(src, dest, width, height);
					if (twidth < (width-2)) twidth += twidth;
					//twidth = (twidth + 1) & 0xfffe;
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureCI4(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
					break;
			case 1:
//					CopyTextureCI8(src, dest, width, height);				
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureCI8(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
					break;
			case 2:  // mhh.. maybe a bug [FiRES]
					CopyTextureCI8(src, dest, width, height);				
					break;

			}			
			break;
	case 3:	//IA
			switch(size)
			{
			case 0:
//					CopyTextureIA4(src, dest, width, height);
					if (twidth < (width-2)) twidth += twidth;
					twidth = (twidth + 1) & 0xfffe;
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureIA4(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
					break;
			case 1:
//					CopyTextureIA8(src, dest, width, height);
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureIA8(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
					break;
			case 2:
//					CopyTextureIA16(src, dest, width, height);
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureIA16(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
					break;
			}			
			break;
	case 4:	//I
            switch(size)
			{
            case 0:
//                    CopyTextureI4(src, dest, width, height);
					if (twidth < (width-2)) twidth += twidth;
					twidth = (twidth + 1) & 0xfffe;
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureI4(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
                    break;
            case 1:
//                    CopyTextureI8(src, dest, width, height);
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureI8(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
                    break;
			case 2:
//                    CopyTextureI16(src, dest, width, height); // Wolken
					src  = &pRDRAM[rdp_reg.TextureImage.addr];
                    NCopyTextureI16(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
                    break;
            }			
            break;
	default:
//			CopyTextureRGBA32(src, dest, width, height);
			src  = &pRDRAM[rdp_reg.TextureImage.addr];
			NCopyTextureRGBA32(src, dest, tmpTile->uls, tmpTile->lrs  , tmpTile->ult, tmpTile->lrt);
			break;
	}
	return 0;
}


int ModTexture(int tTexMode,int Length)
{
	int i; 
	uWORD tmp = 0, tmpClr,tmpClr2;
	int tT0r,tT0g,tT0b,tT0a;
	uWORD *dest = (uWORD *)tbuffer1;

	switch (tTexMode)
	{
	case 0:// no changes
		return 1;
	case AlphaT0:
		for (i=0;i<Length;i++)
		{
			*(dest) |= 0x00ffffff;
			dest++;
		}
		break;
		//AlphaT0Red
	case AlphaT0Red:
		for (i=0;i<Length;i++)
		{
			*(dest+3) = *(dest);
//			*(dest) |= 0x00ffffff;
			dest++;
		}
		break;
	case Tex:
		for (i=0;i<Length;i++)
		{
			*(dest) |= 0xff000000;
			dest++;
		}
		break;
	case Alpha_Mult_Prim:
		tmpClr = ((rdp_reg.primcolor >> 24) & 0xff);
		for (i=0;i<Length;i++)
		{
			tT0a = (*dest & 0xff000000) >> 24;
			tT0a *= tmpClr;
			tT0a >>= 8;
			if (tT0a < 0) tT0a = 0;
			if (tT0a > 0xff) tT0a = 0xff;
			tT0b = *dest & 0xffffff;
			*(dest) = (tT0a<<24) | tT0b;
			dest++;
		}
		break;
	case Alpha_Mult_Env:
		tmpClr = ((rdp_reg.envcolor >> 24) & 0xff);
		for (i=0;i<Length;i++)
		{
			tT0a = (*dest & 0xff000000) >> 24;
			tT0a *= tmpClr;
			tT0a >>= 8;
			if (tT0a < 0) tT0a = 0;
			if (tT0a > 0xff) tT0a = 0xff;
			tT0b = *dest & 0xffffff;
			*(dest) = (tT0a<<24) | tT0b;
			dest++;
		}
		break;
	case Tex | AlphaT0:
		break;
	case Prim_Sub_Tex:
//		tmpClr = ((rdp_reg.primcolor & 0xff) << 24) | ((rdp_reg.primcolor & 0xff00) << 8) | ((rdp_reg.primcolor & 0xff0000) >> 8) | (rdp_reg.primcolor >> 24);
		tmpClr = ((rdp_reg.primcolor & 0xff) << 24) | 
			(((rdp_reg.primcolor & 0xff00) << 16) >> 8) | 
			(((rdp_reg.primcolor & 0xff0000) << 8) >> 16)| 
			((rdp_reg.primcolor & 0xff000000) >> 24);
		for (i=0;i<Length;i++)
		{
			tT0r = (tmpClr & 0xff) - (*dest & 0xff);
			tT0g = (tmpClr & 0xff00) - (*dest & 0xff00);
			tT0b = (tmpClr & 0xff0000) - (*dest & 0xff0000);
			if (tT0r < 0) tT0r = 0;
			if (tT0g < 0) tT0g = 0;
			if (tT0b < 0) tT0b = 0;
			tT0a = *dest & 0xff000000;
			*dest = tT0r | tT0g | tT0b | tT0a;
			dest++;
		}
		break;
	case Env_Sub_Tex:
		tmpClr = ((rdp_reg.envcolor & 0xff) << 24) | ((rdp_reg.envcolor & 0xff00) << 8) | ((rdp_reg.envcolor & 0xff0000) >> 8) | (rdp_reg.envcolor >> 24);
		tmpClr = (tmpClr >> 1) | 0x808080;
		for (i=0;i<Length;i++)
		{
			tT0r = (tmpClr & 0xff) - (*dest & 0xff);
			tT0g = (tmpClr & 0xff00) - (*dest & 0xff00);
			tT0b = (tmpClr & 0xff0000) - (*dest & 0xff0000);
			if (tT0r < 0) tT0r = 0;
			if (tT0g < 0) tT0g = 0;
			if (tT0b < 0) tT0b = 0;
			tT0a = *dest & 0xff000000;
			*dest = tT0r | tT0g | tT0b | tT0a;
			dest++;
		}
		break;
	case Prim_Sub_Tex | Tex:
		tmpClr = ((rdp_reg.primcolor & 0xff) << 24) | ((rdp_reg.primcolor & 0xff00) << 8) | ((rdp_reg.primcolor & 0xff0000) >> 8) | ((rdp_reg.primcolor >> 24) &0xff);
		for (i=0;i<Length;i++)
		{
			tT0r = (tmpClr & 0xff) * (1.0f - ((*dest & 0xff) / 255.0f));
			tT0r += (*dest & 0xff);
			tT0g = (tmpClr & 0xff00) * (1.0f -(((*dest & 0xff00) >> 8) / 255.0f));
			tT0g += (*dest & 0xff00);
			tT0b = (tmpClr & 0xff0000) * (1.0f - (((*dest & 0xff0000) >> 16) / 255.0f));
			tT0b += (*dest & 0xff0000);
			if (tT0r < 0) tT0r = 0;
			if (tT0g < 0) tT0g = 0;
			if (tT0b < 0) tT0b = 0;
			if (tT0r > 0xff) tT0r = (uWORD)0xff;
			if (tT0g > 0xff00) tT0g = (uWORD)0xff00;
			if (tT0b > 0xff0000) tT0b = (uWORD)0xff0000;
			tT0r &= (uWORD)0x0ff;
			tT0g &= (uWORD)0x0ff00;
			tT0b &= (uWORD)0x0ff0000;
			tT0a = *dest & 0xff000000;
			*dest = tT0r | tT0g | tT0b | tT0a;
			dest++;
		}
		break;
	case Tex_Sub_Prim:
//		tmpClr = ((rdp_reg.primcolor & 0xff) << 24) | ((rdp_reg.primcolor & 0xff00) << 8) | ((rdp_reg.primcolor & 0xff0000) >> 8) | (rdp_reg.primcolor >> 24);
		tmpClr = ((rdp_reg.primcolor & 0xff) << 24) | 
			(((rdp_reg.primcolor & 0xff00) << 16) >> 8) | 
			(((rdp_reg.primcolor & 0xff0000) << 8) >> 16)| 
			((rdp_reg.primcolor & 0xff000000) >> 24);
		for (i=0;i<Length;i++)
		{
			tT0r = (*dest & 0xff) - (tmpClr & 0xff);
			tT0g = (*dest & 0xff00) - (tmpClr & 0xff00);
			tT0b = (*dest & 0xff0000) - (tmpClr & 0xff0000);
			if (tT0r < 0) tT0r = 0;
			if (tT0g < 0) tT0g = 0;
			if (tT0b < 0) tT0b = 0;
			tT0a = *dest & 0xff000000;
			*dest = tT0r | tT0g | tT0b | tT0a;
			dest++;
		}
		break;
	case Tex_Sub_Env:
		tmpClr = ((rdp_reg.envcolor & 0xff) << 24) | ((rdp_reg.envcolor & 0xff00) << 8) | ((rdp_reg.envcolor & 0xff0000) >> 8) | (rdp_reg.envcolor >> 24);
		//tmpClr = (tmpClr >> 1) | 0x808080;
		for (i=0;i<Length;i++)
		{
			tT0r = (*dest & 0xff) - (tmpClr & 0xff);
			tT0g = (*dest & 0xff00) - (tmpClr & 0xff00);
			tT0b = (*dest & 0xff0000) - (tmpClr & 0xff0000);
			if (tT0r < 0) tT0r = 0;
			if (tT0g < 0) tT0g = 0;
			if (tT0b < 0) tT0b = 0;
			tT0a = *dest & 0xff000000;
			*dest = tT0r | tT0g | tT0b | tT0a;
			dest++;
		}
		break;
	case Tex_Sub_Env_Mult_Prim_Add_Env:
		tmpClr = ((rdp_reg.envcolor & 0xff) << 24) | 
			((rdp_reg.envcolor & 0xff00) << 8) | 
			((rdp_reg.envcolor & 0xff0000) >> 8) | 
			(rdp_reg.envcolor >> 24);
		tmpClr2 = ((rdp_reg.primcolor & 0xff) << 24) | 
			(((rdp_reg.primcolor & 0xff00) << 16) >> 8) | 
			(((rdp_reg.primcolor & 0xff0000) << 8) >> 16)| 
			((rdp_reg.primcolor & 0xff000000) >> 24);
		for (i=0;i<Length;i++)
		{
			tT0r = ((*dest & 0xff) - (tmpClr & 0xff));
			tT0g = ((*dest & 0xff00) - (tmpClr & 0xff00)) >> 8;
			tT0b = ((*dest & 0xff0000) - (tmpClr & 0xff0000)) >> 16;

			tT0r *= (tmpClr2 & 0xff);
			tT0g *= ((tmpClr2 >>  8) & 0xff);
			tT0b *= ((tmpClr2 >> 16) & 0xff);

			tT0r >>= 8;
			tT0g >>= 8;
			tT0b >>= 8;

			tT0r += (tmpClr & 0xff);
			tT0g += (tmpClr & 0xff00) >> 8;
			tT0b += (tmpClr & 0xff0000) >> 16;

			if (tT0r < 0) tT0r = 0;
			if (tT0g < 0) tT0g = 0;
			if (tT0b < 0) tT0b = 0;

			if (tT0r > 0xff) tT0r = 0xff;
			if (tT0g > 0xff) tT0g = 0xff;
			if (tT0b > 0xff) tT0b = 0xff;

			tT0a = *dest & 0xff000000;
			*dest = tT0r | (tT0g << 8)| (tT0b << 16)| tT0a;
			dest++;
		}
		break;
	case Tex_Sub_Env_Mult_Tex_Add_Prim:
		tmpClr = ((rdp_reg.envcolor & 0xff) << 24) | 
			((rdp_reg.envcolor & 0xff00) << 8) | 
			((rdp_reg.envcolor & 0xff0000) >> 8) | 
			(rdp_reg.envcolor >> 24);
		tmpClr2 = ((rdp_reg.primcolor & 0xff) << 24) | 
			(((rdp_reg.primcolor & 0xff00) << 16) >> 8) | 
			(((rdp_reg.primcolor & 0xff0000) << 8) >> 16)| 
			((rdp_reg.primcolor & 0xff000000) >> 24);
		for (i=0;i<Length;i++)
		{
			tT0r = ((*dest & 0xff) - (tmpClr & 0xff));
			tT0g = ((*dest & 0xff00) - (tmpClr & 0xff00)) >> 8;
			tT0b = ((*dest & 0xff0000) - (tmpClr & 0xff0000)) >> 16;

			tT0r *= ((*dest & 0xff) & 0xff);
			tT0g *= ((*dest >>  8) & 0xff);
			tT0b *= ((*dest >> 16) & 0xff);

			tT0r >>= 8;
			tT0g >>= 8;
			tT0b >>= 8;

			tT0r += (tmpClr2 & 0xff);
			tT0g += (tmpClr2 & 0xff00) >> 8;
			tT0b += (tmpClr2 & 0xff0000) >> 16;

			if (tT0r < 0) tT0r = 0;
			if (tT0g < 0) tT0g = 0;
			if (tT0b < 0) tT0b = 0;

			if (tT0r > 0xff) tT0r = 0xff;
			if (tT0g > 0xff) tT0g = 0xff;
			if (tT0b > 0xff) tT0b = 0xff;

			tT0a = *dest & 0xff000000;
			*dest = tT0r | (tT0g << 8)| (tT0b << 16)| tT0a;
			dest++;
		}
		break;
	case Prim_Sub_Env_Mult_Tex_Add_Env:
		tmpClr = ((rdp_reg.envcolor & 0xff) << 24) | 
			((rdp_reg.envcolor & 0xff00) << 8) | 
			((rdp_reg.envcolor & 0xff0000) >> 8) | 
			(rdp_reg.envcolor >> 24);
		tmpClr2 = ((rdp_reg.primcolor & 0xff) << 24) | 
			(((rdp_reg.primcolor & 0xff00) << 16) >> 8) | 
			(((rdp_reg.primcolor & 0xff0000) << 8) >> 16)| 
			((rdp_reg.primcolor & 0xff000000) >> 24);
		for (i=0;i<Length;i++)
		{
			tmpClr -= tmpClr2;
			tT0r = (tmpClr & 0xff);
			tT0g = (tmpClr & 0xff00) >> 8;
			tT0b = (tmpClr & 0xff0000) >> 16;

			tT0r *= ((*dest & 0xff) & 0xff);
			tT0g *= ((*dest >>  8) & 0xff);
			tT0b *= ((*dest >> 16) & 0xff);

			tT0r >>= 8;
			tT0g >>= 8;
			tT0b >>= 8;

			tT0r += (tmpClr2 & 0xff);
			tT0g += (tmpClr2 & 0xff00) >> 8;
			tT0b += (tmpClr2 & 0xff0000) >> 16;

			if (tT0r < 0) tT0r = 0;
			if (tT0g < 0) tT0g = 0;
			if (tT0b < 0) tT0b = 0;

			if (tT0r > 0xff) tT0r = 0xff;
			if (tT0g > 0xff) tT0g = 0xff;
			if (tT0b > 0xff) tT0b = 0xff;

			tT0a = *dest & 0xff000000;
			*dest = tT0r | (tT0g << 8)| (tT0b << 16)| tT0a;
			dest++;
		}
		break;
	case Env_Sub_Tex_Mult_EnvAlpha_Add_Tex:
		tmpClr = ((rdp_reg.envcolor & 0xff) << 24) | 
			((rdp_reg.envcolor & 0xff00) << 8) | 
			((rdp_reg.envcolor & 0xff0000) >> 8) | 
			(rdp_reg.envcolor >> 24);
		tmpClr2 = rdp_reg.envcolor & 0xff;
		for (i=0;i<Length;i++)
		{

			tT0r = (tmpClr & 0xff);
			tT0g = (tmpClr & 0xff00) >> 8;
			tT0b = (tmpClr & 0xff0000) >> 16;

			tT0r -= ((*dest & 0xff) & 0xff);
			tT0g -= ((*dest >>  8) & 0xff);
			tT0b -= ((*dest >> 16) & 0xff);

			tT0r *= tmpClr2;
			tT0g *= tmpClr2;
			tT0b *= tmpClr2;

			tT0r >>= 8;
			tT0g >>= 8;
			tT0b >>= 8;

			tT0r += ((*dest) & 0xff);
			tT0g += ((*dest >>  8) & 0xff);
			tT0b += ((*dest >> 16) & 0xff);

			if (tT0r < 0) tT0r = 0;
			if (tT0g < 0) tT0g = 0;
			if (tT0b < 0) tT0b = 0;

			if (tT0r > 0xff) tT0r = 0xff;
			if (tT0g > 0xff) tT0g = 0xff;
			if (tT0b > 0xff) tT0b = 0xff;

			tT0a = *dest & 0xff000000;
			*dest = tT0r | (tT0g << 8)| (tT0b << 16)| tT0a;
			dest++;
		}
		break;
	}
	return 1;
}
 
int SetActiveTexture()
{
	uWORD i, new_width, new_height;
	uHWORD *text = &pRDRAM[rdp_reg.TextureImage.addr & 0x3fffff];
	uBYTE *ttext = &pRDRAM[rdp_reg.TextureImage.addr & 0x3fffff];
//** Use m_CurTile now (FiRES)  //**	t_tile *t_p = &rdp_reg.td[0];
	t_tile *t_p = rdp_reg.m_CurTile;
	//t_tile *t_p1 = rdp_reg.m_NexTile;
	sHWORD TexID = -1;
	int offset;
	int NextID;
	int LastID;

	uDWORD addr = rdp_reg.TextureImage.addr;

	int  size = (int)t_p->size;
	int  format = (int)t_p->format;
	int  tult = (int)t_p->ult;
	int  palette = (int)t_p->palette;

	int twidth = t_p->Width;
	int width = rdp_reg.TextureImage.width;

    int stwidth  = t_p->lrs - t_p->uls;
    int stheight = t_p->lrt - t_p->ult;

	if (stwidth < 0) stwidth = -stwidth;
	if (stheight < 0) stheight = -stheight;

    stwidth++;
    stheight++;

//	stheight &= 0x0fffe;

	palette = 0;
	twidth = stwidth;

	if (width <= 1)
		width = twidth;

	if (format == 2)
	{
		switch (size)
		{
		case 0:
			palette += PaletteCRC[t_p->palette];
			break;
		case 1:
			palette += PaletteCRC[16];
			break;
		default:
			palette = 0;
			break;
		}
	}

	switch (size)
	{
	case 0:
		offset = (((tult)* width) + t_p->uls) >> 1;// 4-bit
		break;
	case 1:
		offset =  ((tult) * width) + t_p->uls;// 8-bit
		break;
	case 2:
		offset = (((tult) * width) + t_p->uls) << 1;// 16-bit;
		break;
	case 3:
		offset = (((tult) * width) + t_p->uls) << 2;// 32-bit
		break;
	default:
		offset =  ((tult * width) + t_p->uls);// 8-bit
		break;
	}

//	if (format == 2) offset =  ((tult * width) + t_p->uls);// 8-bit

	ttext = &pRDRAM[rdp_reg.TextureImage.addr + offset];

//	text = ttext;
	TexID = -1;
	NextID = -1;
	LastID = -1;
	if ( (*text == 0xFF00) && (*(text+1) == 0xFE01) && (*(text+2) < num_cached_texture) )
//		TexID = (sHWORD) *(text+2);
		TexID = (sHWORD)*(text+2);
	NextID = TexID;
	if (TexID != -1)
	{
//	for (i=TexID;i<num_cached_texture;i++)
	while (NextID >= 0)
	if (NextID >= 0)
	{
//		if (TexID == index_texture[NextID].TexID)
		{
			if ((offset >> 2) == (index_texture[NextID].offset >> 2))
			{
				if (TexMode == index_texture[NextID].TexMode)
				{
					if (TexColor == index_texture[NextID].TexColor)
					{
						if (palette == index_texture[NextID].Palette)
						{
							glBindTexture(GL_TEXTURE_2D, NextID);
							return 1;
						}
					}
				}
			}
		}
		LastID = NextID;
		NextID = index_texture[NextID].NextID;
	}
	}

//** texture is not cached yet... so we cache it

	index_texture[num_cached_texture].TexID = TexID;
	index_texture[num_cached_texture].offset = offset;
	index_texture[num_cached_texture].TexMode = TexMode;
	index_texture[num_cached_texture].TexColor = TexColor;
	index_texture[num_cached_texture].Palette = palette;
	index_texture[num_cached_texture].NextID = -1;

	glBindTexture(GL_TEXTURE_2D, num_cached_texture);
	
//	if ( (*text == 0xFF00) && (*(text+1) ==0xFE01) && (*(text+2) == i) )
	if ( (*text == 0xFF00) && (*(text+1) ==0xFE01) && (*(text+2) == (uHWORD)TexID) )
	{
//		text = old_texture[i].addr;
		*text = old_texture[TexID].old_bytes[0]; text++;
		*text = old_texture[TexID].old_bytes[1]; text++;
		*text = old_texture[TexID].old_bytes[2];
	}

	DecompressTexture(t_p,ttext,0);

	if (TexID > -1)
	{
		text = old_texture[TexID].addr;
		*text = 0xFF00; text++;
		*text = 0xFE01; text++;
		*text = (uHWORD)TexID;
	}


	if (TexMode != 0) ModTexture(TexMode, t_p->Width * t_p->Height);
	
	text = &pRDRAM[rdp_reg.TextureImage.addr & 0x3fffff];

	if (TexID < 0)
		TexID = (sHWORD)num_cached_texture;

	old_texture[num_cached_texture].addr = text;
	old_texture[num_cached_texture].old_bytes[0] = *text; *text = 0xFF00; text++;
	old_texture[num_cached_texture].old_bytes[1] = *text; *text = 0xFE01; text++;
//	old_texture[num_cached_texture].old_bytes[2] = *text; *text = num_cached_texture;
	old_texture[num_cached_texture].old_bytes[2] = *text; *text = (uHWORD)TexID;

	if (LastID >= 0) 
	{
		index_texture[LastID].NextID = num_cached_texture;
//		MessageBox(hGraphics, "LastID >= 0", "Texture.c", MB_OK);
	}

	if (t_p->clampt) glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
	else glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

	if (t_p->clamps) glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP);
	else glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

//	if (rdp_reg.geometrymode & 0x00080000)
//	{
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
//	}
//	else
//	{
//		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
//		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
//	}

	glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
	
//** Build the texture...	
	if (TestPower2(t_p->Width, t_p->Height, &new_width, &new_height))
	{
		glTexImage2D(GL_TEXTURE_2D, 0, 4, t_p->Width, t_p->Height, 0, GL_RGBA,
				 GL_UNSIGNED_BYTE, (unsigned char*)tbuffer1);
	}
	else
	{
		gluScaleImage(GL_RGBA, t_p->Width, t_p->Height, GL_UNSIGNED_BYTE, (unsigned char*)tbuffer1,
					   new_width, new_height, GL_UNSIGNED_BYTE, (unsigned char*)tbuffer2);

		glTexImage2D(GL_TEXTURE_2D, 0, 4, new_width, new_height, 0, GL_RGBA,
					 GL_UNSIGNED_BYTE, (unsigned char*)tbuffer2);
	}

//** is needed for a lot of demos (FiRES) - if the combines mode work right
//** we dont need this anymore
//	if (NoBlending) glDisable(GL_BLEND);
//	else glEnable(GL_BLEND);

	num_cached_texture++;

// if we have cached to much textures we are clearing the complete texturestuff ...
	if (num_cached_texture >= MAX_CACHED_TEXTURES) 
	{
		num_cached_texture=0;
		for (i=0; i < MAX_CACHED_TEXTURES; i ++)
		{
			text = old_texture[i].addr;
			if ( (*text == 0xFF00) && (*(text+1) ==0xFE01) && (*(text+2) == i) )
			{
				text = old_texture[i].addr;
				*text = old_texture[i].old_bytes[0]; text++;
				*text = old_texture[i].old_bytes[1]; text++;
				*text = old_texture[i].old_bytes[2];
			}
			index_texture[i].TexID		= -1;
			index_texture[i].TexColor	= -1;
			index_texture[i].TexMode	= -1;
			index_texture[i].offset		= -1;
			index_texture[i].NextID		= -1;
			index_texture[i].Palette	= -1;
			glDeleteTextures(1,&i);
		}
		SetActiveTexture();
	}

	TexMode = 0;
	TexColor = 0;
	return 1;
}

/*****************************************************************************/



void CopyTextureRGBA32(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = i_dst;
    uBYTE *src = i_src;
               
    for(i=0; i <= height; i++)
	{
		src = i_src + (i * twidth << 2);
	    for(j=0; j < width; j++)
		{
			src += 3;
            *(dest) = *src; src--; dest++;
            *(dest) = *src; src--; dest++;
            *(dest) = *src; src--; dest++;
            *(dest) = *src; dest++;
            src += 4;
         }
	}


}

static void NCopyTextureRGBA32    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
	uWORD i, j;
	uBYTE *dest = i_dst;
    uBYTE *src = i_src;
               
    for(i=sy; i <= ey; i++)
	{
	    for(j=sx; j <= ex; j++)
		{
			src = i_src + ((i * twidth + j) << 2);
			src += 3;
            *dest = *src; src--; dest++;
            *dest = *src; src--; dest++;
            *dest = *src; src--; dest++;
            *dest = *src; dest++;
            src -= 4;
         }
	}


}

// correct !!
void CopyTextureRGBA16(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
	uWORD i, j;
	uBYTE *dest = i_dst;
    uHWORD *src = (uHWORD*)i_src;
                
    for(i=0; i <= height; i++)
	{
		src = (uHWORD*)i_src + (i * twidth);
	    for(j=0; j < width; j++)
        {
			uWORD  src_l = (uWORD)i_src + ((i * twidth + j) << 1);
			uHWORD *src_h = src_l ^0x02;

			*(dest) = (((*src_h >> 11 ) & 0x001f)<<3); dest++;
			*(dest) = (((*src_h >>  6 ) & 0x001f)<<3); dest++;
			*(dest) = (((*src_h >>  1 ) & 0x001f)<<3); dest++;
			*(dest) =  ((*src_h & 0x01) * 0xff); dest++;
			src++;
         }
	}
}

static void NCopyTextureRGBA16    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
	uWORD i, j;
	uBYTE *dest = i_dst;
    uHWORD *src = (uHWORD*)i_src;
                
    for(i=sy; i <= ey; i++)
	{
		src = (uHWORD*)i_src + (i * twidth);
	    for(j=sx; j <= ex; j++)
        {
			uWORD  src_l = (uWORD)i_src + ((i * twidth + j) << 1);
			uHWORD *src_h = (uHWORD *)(src_l ^0x02);

			*(dest) = (((*src_h >> 11 ) & 0x001f)<<3); dest++;
			*(dest) = (((*src_h >>  6 ) & 0x001f)<<3); dest++;
			*(dest) = (((*src_h >>  1 ) & 0x001f)<<3); dest++;
			*(dest) =  (*src_h & 0x01) ? 0xff : 0x00; dest++;
//			src++;
         }
	}
}

static void CopyTextureCI8(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
   // tested //
        uBYTE   *d = i_dst;
        uBYTE   *s = i_src;
        unsigned int     x, y;
        uHWORD  *pal = (uBYTE *)(pRDRAM + rdp_reg.tlut_8_addr);   /* palette */
//        uHWORD  *pal = (uHWORD *)(&Palette8);   /* palette */
        uHWORD  rgba;   /* color */
		uBYTE index;
		uWORD tsrc;
		uBYTE *tpal;

        for(y = 0; y <= height; y++)
                for(x = 0; x < width; x++)
                {
                    /* 0x800 might be not correct! */
						tsrc = ((uWORD)i_src + (y * twidth) + x) ^ 0x03;
						s = (uBYTE *)tsrc;
						index = *s ^0x01;
						pal = (uBYTE *)(pRDRAM + ((rdp_reg.tlut_8_addr + (index <<1) + 1) ^ 0x1));
                        rgba = *pal; /* load rgba16 from ci8 */
						pal = (uBYTE *)(pRDRAM + ((rdp_reg.tlut_8_addr + (index <<1)) ^ 0x1));
                        rgba |= *pal << 8; /* load rgba16 from ci8 */
                        *d++ = ((rgba & 0xf800) >> 11) << 3;
                        *d++ = ((rgba & 0x07c0) >>  6) << 3;
                        *d++ = ((rgba & 0x003e) >>  1) << 3;
                        *d++ =  (rgba & 0x0001) ? 0xff : 0x00;
                        s++;
                }
}

static void NCopyTextureCI8    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
   // tested //
        uBYTE   *d = i_dst;
        uBYTE   *s = i_src;
        unsigned int     x, y;
        uHWORD  *pal = (uBYTE *)(pRDRAM + rdp_reg.tlut_8_addr);   /* palette */
//        uHWORD  *pal = (uHWORD *)(&Palette8);   /* palette */
        uHWORD  rgba;   /* color */
		uBYTE index;
		uWORD tsrc;
		uBYTE *tpal;

        for(y = sy; y <= ey; y++)
                for(x = sx; x <= ex; x++)
                {
                    /* 0x800 might be not correct! */
						tsrc = ((uWORD)i_src + (y * twidth) + x) ^ 0x03;
						s = (uBYTE *)tsrc;
						index = *s ^0x01;
						pal = (uBYTE *)(pRDRAM + ((rdp_reg.tlut_8_addr + (index <<1) + 1) ^ 0x1));
                        rgba = *pal; /* load rgba16 from ci8 */
						pal = (uBYTE *)(pRDRAM + ((rdp_reg.tlut_8_addr + (index <<1)) ^ 0x1));
                        rgba |= *pal << 8; /* load rgba16 from ci8 */
                        *d++ = ((rgba & 0xf800) >> 11) << 3;
                        *d++ = ((rgba & 0x07c0) >>  6) << 3;
                        *d++ = ((rgba & 0x003e) >>  1) << 3;
                        *d++ =  (rgba & 0x0001) ? 0xff : 0x00;
//                        s++;
                }
}

//        rdp_reg.tlut_4_addr[rdp_reg.td[tile].palette] = (uWORD)rdp_reg.TextureImage.addr;

static void CopyTextureCI4(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
	// TT
    uWORD x, y, tsrc;
    uBYTE *d = i_dst;
    uBYTE *s = i_src;
    uHWORD color;
//    uHWORD *palette = (uHWORD*)(pRDRAM + (int)rdp_reg.tlut_4_addr);
	t_tile *t_p = rdp_reg.m_CurTile;
//    uHWORD *palette = (uHWORD*)(pRDRAM + (int)rdp_reg.tlut_4_addr[t_p->palette]);
    uHWORD *palette = (uHWORD*)((uBYTE*)pRDRAM + (int)rdp_reg.tlut_4_addr[0])+(t_p->palette<<4);
//    uHWORD *palette = (uHWORD*)(pRDRAM + (int)rdp_reg.tlut_8_addr)+(t_p->palette<<4);
    uBYTE index;
	int tw;

	for(y = 0; y < height; y++)
	{
		for(x = 0; x < width; x++)
		{
			/* 0x800 might be not correct! */
			tw = (y * twidth + x) + tstart;
//			if (twidth & 1)
//				tw = ((y<<1) * twidth + x) + tstart;
//			else
//				tw = (y * twidth + x) + tstart;
			tsrc = ((uWORD)i_src + (tw >> 1))^ 0x03;
			s = (uBYTE *)tsrc;

            index = *s ^0x11;
			if (!tw & 1)
			{
				color = palette[index >> 4];
				*d++ = (((color >> 11 ) & 0x001f)<<3);
				*d++ = (((color >>  6 ) & 0x001f)<<3);
				*d++ = (((color >>  1 ) & 0x001f)<<3);
				*d++ = (color & 0x01) ? 0xff : 0x00;
			}
			else
			{
				color = palette[index & 0x0f];
				*d++ = (((color >> 11 ) & 0x001f)<<3);
				*d++ = (((color >>  6 ) & 0x001f)<<3);
				*d++ = (((color >>  1 ) & 0x001f)<<3);
				*d++ = (color & 0x01) ? 0xff : 0x00;
			}
        }
	}
}

static void NCopyTextureCI4    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
	// TT
    uWORD x, y, tsrc;
    uBYTE *d = i_dst;
    uBYTE *s = i_src;
    uHWORD color;
	t_tile *t_p = rdp_reg.m_CurTile;
//    uHWORD *palette = (uHWORD*)(pRDRAM + (int)rdp_reg.tlut_4_addr[t_p->palette]);
//    uHWORD *palette = (uHWORD*)(pRDRAM + (int)rdp_reg.tlut_8_addr);
//    uHWORD *palette = (uHWORD*)((pRDRAM + (int)rdp_reg.tlut_8_addr))+(t_p->palette<<4);
//	uHWORD *palette = ((uHWORD*)((pRDRAM + (int)rdp_reg.tlut_8_addr)))+(t_p->palette<<4);
        uHWORD  *palette = (uHWORD *)(&Palette8)+(t_p->palette<<4);   /* palette */

    uBYTE index;
	int tw;

	for(y = sy; y <= ey; y++)
	{
		for(x = sx; x <= ex; x++)
		{
			tw = (y * twidth + x);
			tsrc = ((uWORD)i_src + (tw >> 1))^ 0x03;
			s = (uBYTE *)tsrc;

            index = *s ^0x11;
//			if (!tw & 1)
			if ((x & 1) == 0)
			{
				color = palette[index >> 4];
				*d++ = (((color >> 11 ) & 0x001f)<<3);
				*d++ = (((color >>  6 ) & 0x001f)<<3);
				*d++ = (((color >>  1 ) & 0x001f)<<3);
				*d++ = (color & 0x01) ? 0xff : 0x00;
			}
			else
			{
				color = palette[index & 0x0f];
				*d++ = (((color >> 11 ) & 0x001f)<<3);
				*d++ = (((color >>  6 ) & 0x001f)<<3);
				*d++ = (((color >>  1 ) & 0x001f)<<3);
				*d++ = (color & 0x01) ? 0xff : 0x00;
			}
        }
	}
}

void writetex()
{
	unsigned int i, j = 0;
	FILE *stream = fopen("d:/texture.raw","wb");
	for (i = 0; i < rdp_reg.td[0].Width*rdp_reg.td[0].Height; i++)
	{
		fprintf(stream,"%c",tbuffer1[j]); j++;
		fprintf(stream,"%c",tbuffer1[j]); j++;
		fprintf(stream,"%c",tbuffer1[j]); j++;
		j++;
	}
	fclose(stream);	
}

static void CopyTextureIA16(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
	// TT
	uWORD x, y, tsrc;
	uBYTE *dest = i_dst;
    uBYTE *src = i_src;
	uBYTE intensity;
                
    for(y=0; y <= height; y++)
	{
//		src = i_src + (y * twidth<<1);
	    for(x=0; x < (width); x+=1)
        {
			tsrc = ((uWORD)i_src + ((y * twidth + x) << 1)) ^ 0x02;
			src = (uBYTE *)tsrc;
			intensity = *(src + 1);
			*(dest+4) =  intensity;
			*(dest+5) =  intensity;
			*(dest+6) =  intensity;
 	  	    *(dest+7) = *(src + 0);

			dest+=4;
        }
	}
}

static void NCopyTextureIA16   (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
	// TT
	uWORD x, y, tsrc;
	uBYTE *dest = i_dst;
    uBYTE *src = i_src;
	uBYTE intensity;
                
    for(y=sy; y <= ey; y++)
	{
	    for(x=sx; x <= ex; x++)
        {
			tsrc = ((uWORD)i_src + (((y * twidth) + x) << 1)) ^ 0x02;
			src = (uBYTE *)tsrc;
			intensity = *(src + 1);
			*dest =  intensity; dest++;
			*dest =  intensity; dest++;
			*dest =  intensity; dest++;
 	  	    *dest = *(src + 0); dest++;
        }
	}
}

static void CopyTextureIA8(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
	// TT
	uWORD x,y,i,tsrc;
	uBYTE *dest = i_dst;
    uBYTE *src = i_src;
//    float r,g,b,a;
	
    for(y=0; y < height; y++)
	    for(x=0; x < width; x++)
        {
			tsrc = ((uWORD)i_src + (y * twidth) + x) ^ 0x03;
			src = (uBYTE *)tsrc;
			i = (*src &0xf0);
			i |= i >> 4;
			*dest = i; dest++;
			*dest = i; dest++;
			*dest = i; dest++;
 	  	    *dest = (*src & 0x0f) | ((*src & 0x0f) << 4); dest++; // aplha
        }

}

static void NCopyTextureIA8   (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
	// TT
	uWORD x,y,i,tsrc;
	uBYTE *dest = i_dst;
    uBYTE *src = i_src;
//    float r,g,b,a;
	
    for(y=sy; y <= ey; y++)
	    for(x=sx; x <= ex; x++)
        {
			tsrc = ((uWORD)i_src + (y * twidth) + x) ^ 0x03;
			src = (uBYTE *)tsrc;
			i = (*src &0xf0);
			i |= i >> 4;
			*dest = i; dest++;
			*dest = i; dest++;
			*dest = i; dest++;
 	  	    *dest = (*src & 0x0f) | ((*src & 0x0f) << 4); dest++; // aplha
        }

}

static void CopyTextureIA4(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
	// TT
    uWORD x, y, tsrc;
    uBYTE *d = i_dst;
    uBYTE *s = i_src;
    uBYTE i;
	int tw;

	for(y = 0; y < height; y++)
	{
		for(x = 0; x < width; x++)
		{
			/* 0x800 might be not correct! */
			if (twidth & 1)
				tw = ((y<<1) * twidth + x) + tstart;
			else
				tw = (y * twidth + x) + tstart;
			tsrc = ((uWORD)i_src + (tw >> 1))^ 0x03;
			s = (uBYTE *)tsrc;

			if (!tw & 1)
			{
					i = (*s & 0xf0) | ((*s & 0xf0) >> 4);
					*d++ = i;
					*d++ = i;
					*d++ = i;
					*d++ = (i & 1) ? 0xff : 0x00;
			}
			else
			{
					i = (*s & 0x0f) | ((*s & 0x0f) << 4);
					*d++ = i;
					*d++ = i;
					*d++ = i;
					*d++ = (i & 1) ? 0xff : 0x00;
			}
        }
		//y++;
	}
}

static void NCopyTextureIA4    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
	// TT
    uWORD x, y, tsrc;
    uBYTE *d = i_dst;
    uBYTE *s = i_src;
    uBYTE i;
	int tw;

	for(y = sy; y <= ey; y++)
	{
		for(x = sx; x <= ex; x++)
		{
			/* 0x800 might be not correct! */
			tw = (y * twidth) + x;
			tsrc = (uWORD)i_src + ((tw >> 1)^ 0x03);
			s = (uBYTE *)tsrc;

//			if (!tw & 1)
			if ((x & 1) == 0)
			{
					i = (*s & 0xf0) | ((*s & 0xf0) >> 4);
					*d++ = i;
					*d++ = i;
					*d++ = i;
					*d++ = (i & 1) ? 0xff : 0x00;
			}
			else
			{
					i = (*s & 0x0f) | ((*s & 0x0f) << 4);
					*d++ = i;
					*d++ = i;
					*d++ = i;
					*d++ = (i & 1) ? 0xff : 0x00;
			}
        }
		//y++;
	}
}

static void CopyTextureI16(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
	uWORD x, y, tsrc;
	uBYTE *dest = i_dst;
    uBYTE *src = i_src;
	uBYTE intensity;
                
    for(y=0; y < height; y++)
	{
		src = i_src + (y * twidth<<1);
	    for(x=0; x < (width); x+=1)
        {
			tsrc = ((uWORD)i_src + ((y * twidth + x) << 1)) ^ 0x02;
			src = (uBYTE *)tsrc;
			intensity = *(src + 1);
			*(dest+0) =  intensity;
			*(dest+1) =  intensity;
			*(dest+2) =  intensity;
 	  	    *(dest+3) = *(src + 1);
			dest+=4;
        }
	}
}

static void NCopyTextureI16    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
	uWORD x, y, tsrc;
	uBYTE *dest = i_dst;
    uBYTE *src = i_src;
	uBYTE intensity;
                
    for(y=sy; y <= ey; y++)
	{
		src = i_src + (y * twidth<<1);
	    for(x=sx; x <= ex; x++)
        {
			tsrc = ((uWORD)i_src + (((y * twidth) + x) << 1)) ^ 0x02;
			src = (uBYTE *)tsrc;
			intensity = *(src + 1);
			*dest =  intensity; dest++;
			*dest =  intensity; dest++;
			*dest =  intensity; dest++;
 	  	    *dest = *(src + 1); dest++;
        }
	}
}

static void CopyTextureI8(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
        uWORD  x, y, tsrc;
        uBYTE  *d = i_dst;
        uBYTE  *s = i_src;
        uBYTE  i;

        for(y=0; y < height; y++)
                for(x=0; x < width; x++)
                {
                    s = i_src + (y * twidth) + (x ^ 0x3);    
//					tsrc = ((uWORD)i_src + (y * twidth) + x) ^ 0x03;
//					s = (uBYTE *)tsrc;
					i = *s;
					*d++ = i;
                    *d++ = i;
                    *d++ = i;
                    *d++ = i; 
                    //s++;
                }
}

static void NCopyTextureI8    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
        uWORD  x, y, tsrc;
        uBYTE  *d = i_dst;
        uBYTE  *s = i_src;
        uBYTE  i;

        for(y=sy; y <= ey; y++)
                for(x=sx; x <= ex; x++)
                {
                    s = i_src + (((y * twidth) + x) ^ 0x3);    
//					tsrc = ((uWORD)i_src + (y * twidth) + x) ^ 0x03;
//					s = (uBYTE *)tsrc;
					i = *s;
					*d++ = i;
                    *d++ = i;
                    *d++ = i;
                    *d++ = i; 
                    //s++;
                }
}

static void CopyTextureI4(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
	// TT
    uWORD x, y, tsrc;
    uBYTE *d = i_dst;
    uBYTE *s = i_src;
    uBYTE i;
	int tw;

	for(y = 0; y < height; y++)
	{
		for(x = 0; x < width; x++)
		{
			/* 0x800 might be not correct! */
			if (width & 1)
				tw = ((y<<1) * twidth + x) + tstart;
			else
				tw = (y * twidth + x) + tstart;
			tsrc = ((uWORD)i_src + (tw >> 1))^ 0x03;
			s = (uBYTE *)tsrc;

			if (!tw & 1)
			{
					i = (*s & 0xf0) | ((*s & 0xf0) >> 4);
					*d++ = i;
					*d++ = i;
					*d++ = i;
					*d++ = i;
			}
			else
			{
					i = (*s & 0x0f) | ((*s & 0x0f) << 4);
					*d++ = i;
					*d++ = i;
					*d++ = i;
					*d++ = i;
			}
        }
		//y++;
	}
}

static void NCopyTextureI4    (uBYTE *i_src, uBYTE *i_dst, uWORD sx, uWORD ex, uWORD sy, uWORD ey)
{
	// TT
    uWORD x, y, tsrc;
    uBYTE *d = i_dst;
    uBYTE *s = i_src;
    uBYTE i;
	int tw;

	for(y = sy; y <= ey; y++)
	{
		for(x = sx; x <= ex; x++)
		{
			/* 0x800 might be not correct! */
			tw = (y * twidth) + x;
			tsrc = (uWORD)i_src + ((tw >> 1)^ 0x03);
			s = (uBYTE *)tsrc;

//			if (!tw & 1)
			if ((x & 1) == 0)
			{
					i = (*s & 0xf0) | ((*s & 0xf0) >> 4);
					*d++ = i;
					*d++ = i;
					*d++ = i;
					*d++ = i;
			}
			else
			{
					i = (*s & 0x0f) | ((*s & 0x0f) << 4);
					*d++ = i;
					*d++ = i;
					*d++ = i;
					*d++ = i;
			}
        }
		//y++;
	}
}


/*****************************************************************************/
/*** the follow function is  not implemented yet *****************************/
/*****************************************************************************/

static void CopyTextureYUV(uBYTE *i_src, uBYTE *i_dst, uWORD width, uWORD height)
{
    uWORD i, j;
    uBYTE *d = i_dst;
    uHWORD *s = (uHWORD*)i_src;

    for(i=0; i < height; i++)
        for(j=0; j < width; j++)
        {
            *d++ = 100;
            *d++ = 150;
            *d++ = 200;
            *d++ = 255;
            s++;
        }

}
