/**
 * TrueReality - N64/rdp_registers.h
 * Copyright (C) 1998, 1999 Niki W. Waibel
 *
 * This program is free software; you can redistribute it and/
 * or modify it under the terms of the GNU General Public Li-
 * cence as published by the Free Software Foundation; either
 * version 2 of the Licence, or any later version.
 *
 * This program is distributed in the hope that it will be use-
 * ful, but WITHOUT ANY WARRANTY; without even the implied war-
 * ranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public Licence for more details.
 *
 * You should have received a copy of the GNU General Public
 * Licence along with this program; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,
 * USA.
 *
 * Information about me (the author):
 *   Niki W. Waibel, Reichenau 20, 6890 Lustenau, Austria - EUROPE
 *   niki.waibel@gmx.net
**/





/* every 'export' in this file resides in 'rdp.c' */
/* this file could also be named as 'rdp_export.h' */




#ifndef __RDP_REGISTERS_H
#define __RDP_REGISTERS_H

#include "type_sizes.h"

#define DEFAULT						0x00
#define USE_COLOR4_PRIM				0x01
#define USE_COLOR4_VERTEX_PRIM		0x02
#define USE_COLOR3_VERTEX			0x04
#define USE_TEXTURE					0x10
#define USE_ALPHA					0x20
#define USE_FREE					0x40

//**  structure vertex
typedef struct
{
	float   x, y, z;
    uHWORD   flags;
    float   s, t;
    uBYTE    r, g, b, a;
}
t_vtx;                           

typedef struct
{
	float    x, y, z;
    uHWORD   flags;
    float    s, t;
    sBYTE    nx, ny, nz;
	uBYTE	 a;
}
t_vtx_tn;

typedef struct
{
	float   sn1, cs1;
	float   sn2, cs2;
    float    r, g, b, a;
}
t_lightsw;                           

typedef struct
{
	float r, g, b, a;
	float r_copy, g_copy, b_copy, a_copy;
	float x, y, z, w;
} t_light;

//** first by Jeff
typedef struct
{
	uBYTE  fmt;			// frame format (3 bits)
	uBYTE  size;		// bits per pixel (2 bits)
	uWORD  width;		// width of frame in pixels (12 bits)
	uDWORD addr;		// RDRAM pointer to Image
	uBYTE *LoadAddr;
	uDWORD vaddr;
} t_ImageInfo;

typedef struct
{
//** SETTILE 
	uBYTE    format;
	uBYTE    size;
	uHWORD   line;
	uHWORD   tmem;
	uBYTE    palette;
	uBIT     clampt;
	uBIT     mirrort;
	uBYTE    maskt;
	uBYTE    shiftt;
	uBIT     clamps;
	uBIT     mirrors;
	uBYTE    masks;
	uBYTE    shifts;

	uWORD	 Width;
	uWORD	 Height;
	
	uBYTE    Texture_on;
	float	 SScale;
	float	 TScale;

#define RDPTD_LOAD_NOT_SET -1
#define RDPTD_LOADBLOCK 1
#define RDPTD_LOADTILE 2
	int		set_by;
	uWORD   addr;
//** LOADBLOCK / LOADTILE
	uHWORD   uls; /* Upper Left Texture X-Coordinate (10.2 bits) */
	uHWORD   ult; /* Upper Left Texture Y-Coordinate (10.2 bits) */
	uHWORD   lrs; /* Lower Right Texture X-Coordinate (10.2 bits) */
	uHWORD   lrt; /* Lower Right Texture Y-Coordinate (10.2 bits) */
	uHWORD   dxt; /* Increment dxt per scanline */

//**  SETTILESIZE
	float   sts_uls; /* Upper Left Texture X-Coordinate (10.2 bits) */
	float   sts_ult; /* Upper Left Texture Y-Coordinate (10.2 bits) */
	float   sts_lrs; /* Lower Right Texture X-Coordinate (10.2 bits) */
	float   sts_lrt; /* Lower Right Texture Y-Coordinate (10.2 bits) */
	
} t_tile;


typedef struct
{
	uWORD    ulx;
	uWORD    uly;
	uWORD    lrx;
    uWORD    lry;
    uWORD    interlace_mode;
} t_scissor;

typedef struct 
{
        //** general 

        uWORD  cmd0, cmd1;
        uWORD  start, stop, curr;

        t_tile  td[8];
		int		tile;
		int		loadtile;

        //** color register 
        uWORD  fillcolor;     //** memory interface: 32bit 
        uWORD  primcolor;     //** color combiner: rgba 
        uWORD  envcolor;      //** color combiner: rgba 
        uWORD  fogcolor;      //** blender: rgba 
        uWORD  blendcolor;    //** blender: rgba 

        //** This is the frame buffer location to draw into 
        //** Used by rdp_setcolorimage() 
        uBYTE  colorimg_fmt;          //** Frame Format (3 bits)
        uBYTE  colorimg_size;         //** Bits per pixel (2 bits) 
        uHWORD colorimg_width;        //** width of frame in pixels (12 bits) 
        uWORD  colorimg_addr;         //** RDRAM pointer to Image (32 bits) 

        //** This is the frame buffer location to draw into 
        //** Used by rdp_settextureimage() 
		t_ImageInfo TextureImage;
//** FiRES: muesste theoretisch fuer jedes der geladenen Texturen indiv.
		float m_TScale;				
		float m_SScale;
		t_tile *m_CurTile;

        //** This is the location of an 8 bit TLUT 
        //** Used by rdp_loadtlut() 
        uBYTE  tlut_8_fmt;   //** TLUT Format (3 bits) 
        uBYTE  tlut_8_size;  //** Bits per pixel (2 bits) 
        uWORD  tlut_8_addr;  //** RDRAM pointer to TLUT (32 bits) 

        //** This is the location of an 4 bit TLUT 
        //** Used by rdp_loadtlut() 
        uBYTE  tlut_4_fmt[16];   //** TLUT Format (3 bits) 
        uBYTE  tlut_4_size[16];  //** Bits per pixel (2 bits) 
        uWORD  tlut_4_addr[16];  //** RDRAM pointer to TLUT (32 bits) 

        t_scissor scissor;


		t_light light[8], lookat_x, lookat_y;


        /**
        *
        * This might be part of some sort of High Level Emulation (HLE).
        * HLE was first hacked out by Epsilon and RealityMan - the authors of UltraHLE.
        *
        **/

        uWORD pc[10];                            /* DL stack is 10 */
        uWORD pc_i;                              /* PC index (current DL depth) */

#define INVALID_TASK    -1
#define GFX_TASK        1
#define AUDIO_TASK      2

        int task;
        int size;   /* length of task */



        struct
        {
                uWORD mipmap_level;
                uWORD on;               /* 1: on 0: off */
                float scale_t;
                float scale_s;
        }
        texture[8];

        uWORD geometrymode;
		uWORD geometrymode_textures;
        uWORD mode_h;
        uWORD mode_l;
        uWORD lights;
        uWORD segment[16];                       /* segment pointer */

        float vp[8];                            /* viewport (scale, translate) */

        /*
        float perspective_normal;
        */

        float mtx_proj[4][4];                   /* projection matrix */

        float mtx_model[10][4][4];              /* 10 modelview matrices */

        t_vtx vtx[256];                         /* 32 vertices */

        struct
        {
                sWORD   nx;
                sWORD   ny;
                sWORD   px;
                sWORD   py;
        }
        clip;                                   /* clipping stuff */


        int halt;
		int useLights;
		int graphics_changes;
//		t_tile *m_NexTile;
//		int useT1;
        uWORD  cmd2, cmd3;
} t_rdp_reg; /* struct rsp_reg */

extern t_rdp_reg rdp_reg;
extern int PaletteCRC[17];

#endif